/**
 * Takes a list and returns a list of lists where each sublist's elements are
 * all satisfied pairwise comparison according to the provided function.
 *
 * @func
 * @memberOf R
 * @since v0.21.0
 * @category List
 * @sig ((a, a) → Boolean) → [a] → [[a]]
 * @param {Function} fn Function for determining whether two given (adjacent)
 *        elements should be in the same group
 * @param {Array} list The array to group. Also accepts a string, which will be
 *        treated as a list of characters.
 * @return {List} A list that contains sublists of elements,
 *         whose concatenations are equal to the original list.
 * @example
 *
 * R.groupWith(R.equals, [0, 1, 1, 2, 3, 5, 8, 13, 21])
 * //=> [[0], [1, 1], [2], [3], [5], [8], [13], [21]]
 *
 * R.groupWith((a, b) => a + 1 === b, [0, 1, 1, 2, 3, 5, 8, 13, 21])
 * //=> [[0, 1], [1, 2, 3], [5], [8], [13], [21]]
 *
 * R.groupWith((a, b) => a % 2 === b % 2, [0, 1, 1, 2, 3, 5, 8, 13, 21])
 * //=> [[0], [1, 1], [2], [3, 5], [8], [13, 21]]
 *
 * R.groupWith(R.eqBy(isVowel), 'aestiou')
 * //=> ['ae', 'st', 'iou']
 */
export function groupWith(predicate, list){
  const toReturn = []
  let holder = []

  list.reduce((prev, current, i) => {
    if (i > 0 && predicate(prev, current)){
      if (holder.length === 0){
        holder.push(prev)
        holder.push(current)
      } else {
        holder.push(current)
      }
    } else if (i > 0){
      if (holder.length === 0){
        toReturn.push([ prev ])
        if (i === list.length - 1) holder.push(current)
      } else {
        toReturn.push(holder)
        holder = []
      }
    }

    return current
  }, undefined)

  return holder.length === 0 ? toReturn : [ ...toReturn, holder ]
}
